#include "GainBucket.h"

struct bucketentry *BucketInsert(struct gainbucket *pGB, long key, long VtxNr)
/* Inserts vertex number in gainbucket with key value. 
   Vertex must not be present already. */
{
  struct bucket *pB, **ppB ;
  struct bucketentry *pE ;
  
  ppB = &(pGB->Root) ;
  pB = *ppB ;

  /*## While the key is smaller than the value in the buckets, 
       go down the bucket list: ##*/
  while ( *ppB != NULL ) 
    {
      pB = *ppB ;

      if ( key < pB->value )
	ppB = &(pB->next) ;
      else 
	break ;
    }

  if ( pB == NULL )
    { /*## Bucket list is empty. Create first bucket: ##*/

      *ppB = (struct bucket *) malloc(sizeof(struct bucket)) ;
      if ( *ppB == NULL )
	Abort("BucketInsert(): not enough memory for first bucket") ;
      
      (*ppB)->value = key ;
      (*ppB)->entry = NULL ;
      (*ppB)->prev = (*ppB)->next = NULL ;

      pGB->NrBuckets++ ;
    }
  else 
    if ( key < pB->value ) 
      { /*## Create new bucket at the end of the current list: ##*/
        if ( *ppB != NULL )
          Abort("BucketInsert(): Internal error") ;

        *ppB = (struct bucket *) malloc(sizeof(struct bucket)) ;
	if ( *ppB == NULL )
	  Abort("BucketInsert(): not enough memory for new bucket") ;

	(*ppB)->value = key ;
	(*ppB)->entry = NULL ;
	(*ppB)->prev = pB ;
	(*ppB)->next = NULL ;

	pGB->NrBuckets++ ;	  
      }
    else
      if ( key > pB->value ) 
        { /*## Create new bucket between the current and the previous 
	       bucket in the list: ##*/

	  *ppB = (struct bucket *) malloc(sizeof(struct bucket)) ;
	  if ( *ppB == NULL )
            Abort("BucketInsert(): not enough memory for new bucket") ;
	    
	  (*ppB)->value = key ;
	  (*ppB)->entry = NULL ;
	  (*ppB)->prev = pB->prev ;
	  (*ppB)->next = pB ;
	   
	  if ( pB->prev != NULL )
	    (pB->prev)->next = *ppB ;
	  else
	    pGB->Root = *ppB ;
	    
	  pB->prev = *ppB ;
	  pGB->NrBuckets++ ;
	}
      /* otherwise key == pB->value */


  /*## Insert the vertex number into the bucketentry: ##*/
  pE = (struct bucketentry *) malloc(sizeof(struct bucketentry)) ;
  if ( pE == NULL )
    Abort("BucketInsert(): not enough memory for bucket entry") ;

  pE->vtxnr = VtxNr ;
  pE->bucket = *ppB ;  
  pE->prev = NULL ;
  pE->next = (*ppB)->entry ;
  if ( pE->next != NULL )
    (pE->next)->prev = pE ;
  (*ppB)->entry = pE ;

  return(pE) ;
}


struct bucketentry *BucketMove(struct gainbucket *pGB, struct bucketentry *pE,
			       long key) 
/* Moves vertex to gainbucket with key value. 
   The key value must differ from the current gain value. */
{
  long VtxNr ;
  struct bucket *pB ;

  VtxNr = pE->vtxnr ;
  pB = pE->bucket ;

  if ( pB->value == key )
    Abort("BucketMove(): destination is equal to origin") ;

  /*## Remove this entry from the bucket: ##*/
  if ( pE->prev != NULL )
    (pE->prev)->next = pE->next ;  
  else
    pB->entry = pE->next ;
  if ( pE->next != NULL )
    (pE->next)->prev = pE->prev ;
  if ( pE != NULL ) 
    free(pE) ;

  /*## If this bucket is now empty, remove it from the list: ##*/
  if ( pB->entry == NULL )
    {      
      if ( pB->prev != NULL )
	(pB->prev)->next = pB->next ;
      else
	pGB->Root = pB->next ;

      if ( pB->next != NULL )
	(pB->next)->prev = pB->prev ;
      
      if ( pB != NULL )
        free(pB) ;
      pGB->NrBuckets-- ;
    }
  
  /*## Insert the vertex in the bucket of the new key: ##*/
  pE = BucketInsert(pGB, key, VtxNr) ;

  return(pE) ;
}

void PrintBuckets(struct gainbucket GB)
{
  long nrentries = 0 ;
  struct bucket *pB ;
  struct bucketentry *pE ;

  printf("Number of Buckets: %ld \n", GB.NrBuckets) ;
  
  pB = GB.Root ;

  while ( pB != NULL )
    {
      pE = pB->entry ;
      while ( pE != NULL )
	{
	  printf("B[% ld] vntnr: %ld \n", (pE->bucket)->value, pE->vtxnr) ;
	  nrentries++ ;
	  pE = pE->next ;
	}
      pB = pB->next ;
    }

  printf("TotalNrEntries: %ld\n", nrentries) ;

}


long GainBucketExtractMax(struct gainbucket *pGB) 
{
  long VtxNr ;
  struct bucket *pB ;
  struct bucketentry *pE ;
  
  pB = pGB->Root  ;
  if ( pB == NULL )
    Abort("GainBucketExtractMax(): Internal error") ;

  pE = pB->entry ;  
  if ( pE == NULL )
    Abort("GainBucketExtractMax(): Internal error") ;

  VtxNr = pE->vtxnr ;

  /*## Remove this entry from the bucket: ##*/
  pB->entry = pE->next ;
 
  if ( pE->next != NULL )
    (pE->next)->prev = NULL ;

  free(pE) ;

  /*## If this bucket is now empty, remove it from the list: ##*/
  if ( pB->entry == NULL )
    {      
      pGB->Root = pB->next ;

      if ( pB->next != NULL )
	(pB->next)->prev = NULL ;
      
      free(pB) ;
      pGB->NrBuckets-- ;
    }

  return(VtxNr) ;
}

long GainBucketGetMaxVal(struct gainbucket *pGB) 
{
  if ( pGB->NrBuckets > 0 )
    return((pGB->Root)->value) ;
  else
    return(LONG_MIN) ;
}

long GainBucketGetMaxValVertexNr(struct gainbucket *pGB) 
{
  if ( pGB->NrBuckets > 0 )
    return(((pGB->Root)->entry)->vtxnr) ;
  else
    return(LONG_MIN) ;
}

void ClearGainBucket(struct gainbucket *pGB) 
{
  struct bucket *pB ;
  struct bucketentry *pE ;
  
  while ( (pB = pGB->Root) != NULL )
    {      
      pGB->Root = pB->next ;

      /*## Remove all entries from this bucket: ##*/  
      while ( (pE = pB->entry) != NULL )
	{
	  pB->entry = pE->next ;
	  free(pE) ;
	}

      /*## Remove this bucket from the list: ##*/
      free(pB) ;
      pGB->NrBuckets-- ;
    }
}
