#include "Graph.h"
#include "HKLFM.h"

extern struct opts Options ;

struct biparthypergraph CreateNewBiPartHyperGraph(long NrVertices, long NrNets,
		long NrEdges, short StoreMat, char MType)
{ 
  register long t ;
  struct biparthypergraph HG ;

  HG.NrVertices = NrVertices ;
  HG.NrNets = NrNets ;
  HG.NrEdges = NrEdges ;
  
  /*## Allocate memory for the vertices: ##*/
  if ( HG.NrVertices > 0 )
    {
      HG.V = (struct vertex *) malloc(HG.NrVertices * sizeof(struct vertex)) ;
      HG.VtxMoveLog = (long *) malloc(HG.NrVertices * sizeof(long)) ;
      HG.OptPartVtx = (int *) malloc(HG.NrVertices * sizeof(int)) ;

      if ( HG.V == NULL || HG.VtxMoveLog == NULL || HG.OptPartVtx == NULL )
	Abort("CreateNewBiPartHyperGraph(): "
	      "not enough memory for the vertices") ;

      if ( StoreMat == TRUE )
	{
	  HG.Vtx2MatIndex = (long *) malloc(HG.NrVertices * sizeof(long)) ;
	  
	  if ( HG.Vtx2MatIndex == NULL )
	    Abort("CreateNewBiPartHyperGraph(): "
		  "not enough memory for the vertices") ;
	}
      else
	HG.Vtx2MatIndex = NULL ;
    }

  /*## Allocate memory for the nets: ##*/
  if ( HG.NrNets > 0 )
    {
      HG.N = (struct net *) malloc(HG.NrNets * sizeof(struct net)) ;

      if ( HG.N == NULL )
	Abort("CreateNewBiPartHyperGraph(): "
	      "not enough memory for the nets") ;

      if ( StoreMat == TRUE )
	{
	  HG.Net2MatIndex = (long *) malloc(HG.NrNets * sizeof(long)) ;
	  
	  if ( HG.Net2MatIndex == NULL )
	    Abort("CreateNewBiPartHyperGraph(): "
		  "not enough memory for the nets") ;
	}
      else
	HG.Net2MatIndex = NULL ;
    }

  /*## Allocate memory for the edges: ##*/
  HG.MatReValue = NULL ;
  HG.MatImValue = NULL ;

  if ( HG.NrEdges > 0 )
    {
      HG.VtxAdjncy = (long *) malloc(HG.NrEdges * sizeof(long)) ;
      HG.NetAdjncy = (long *) malloc(HG.NrEdges * sizeof(long)) ;
      HG.EdgeWeight = (long *) malloc(HG.NrEdges * sizeof(long)) ;      
      
      if ( HG.VtxAdjncy == NULL || HG.NetAdjncy == NULL || 
	   HG.EdgeWeight == NULL )  
	Abort("CreateNewBiPartHyperGraph(): "
	      "not enough memory for the edges") ;

      if ( StoreMat == TRUE )
	{
          if (MType != 'P') 
            {
	      HG.MatReValue = (double *) malloc(HG.NrEdges * sizeof(double)) ;
	  
	      if ( HG.MatReValue == NULL )
	        Abort("CreateNewBiPartHyperGraph(): "
		      "not enough memory for the edges") ;
            }
          if (MType == 'C') 
            {
	      HG.MatImValue = (double *) malloc(HG.NrEdges * sizeof(double)) ;
	  
	      if ( HG.MatImValue == NULL )
	        Abort("CreateNewBiPartHyperGraph(): "
		      "not enough memory for the edges") ;
            }
	}
    }

  /*## Clear all values for the vertices: ##*/
  for ( t = 0 ; t < HG.NrVertices ; t++ )
    {
      HG.V[t].vtxwgt = 0 ;
      HG.V[t].iStart = 0 ;
      HG.V[t].iEnd = 0 ;
      HG.V[t].partition = 0 ;
      HG.V[t].GBentry = NULL ;      
      HG.OptPartVtx[t] = -1 ;
      if ( StoreMat == TRUE )
	HG.Vtx2MatIndex[t] = 0 ;
    }

  /*## Clear all values for the nets: ##*/
  for ( t = 0 ; t < HG.NrNets ; t++ )
    {
      HG.N[t].netwgt = 0 ;
      HG.N[t].iStartP0 = 0 ;
      HG.N[t].iStartP1 = 0 ;
      HG.N[t].iEnd = 0 ;
      if ( StoreMat == TRUE )
	HG.Net2MatIndex[t] = 0 ;
    }

  /*## Clear all values for the edges: ##*/
  for ( t = 0 ; t < HG.NrEdges ; t++ )
    {
      HG.VtxAdjncy[t] = 0 ;
      HG.NetAdjncy[t] = 0 ;
      HG.EdgeWeight[t] = 0 ;
      if ( StoreMat == TRUE )
        {
          if (MType != 'P')
	    HG.MatReValue[t] = 0.0 ;
          if (MType == 'C')
	    HG.MatImValue[t] = 0.0 ;
        }
    }

  /*## Initialize the gain buckets: ##*/
  HG.GBVtx[0].NrBuckets = HG.GBVtx[1].NrBuckets = 0 ;
  HG.GBVtx[0].Root = HG.GBVtx[1].Root = NULL ;

  /*## Clear weights: ##*/
  HG.WeightP[0] = HG.WeightP[1] = 0 ;

  /*## Clear comm / log / dir: ##*/
  HG.CurComm = HG.MinComm = HG.OptComm = 0 ;
  HG.CurVtxLog = HG.MinVtxLog = 0 ;
  HG.SplitDir = ROW ;

  return(HG) ;
}

void DeleteBiPartHyperGraph(struct biparthypergraph HG)
{
  if ( HG.NrVertices > 0 )
    {
      if ( HG.V != NULL )
	free(HG.V) ;      
      if ( HG.VtxMoveLog != NULL )
	free(HG.VtxMoveLog) ;
      if ( HG.OptPartVtx != NULL )
	free(HG.OptPartVtx) ;
      if ( HG.Vtx2MatIndex != NULL )
	free(HG.Vtx2MatIndex) ;
    }

  if ( HG.NrNets > 0 )
    {
      if ( HG.N != NULL )
	free(HG.N) ;
      if ( HG.Net2MatIndex != NULL )
	free(HG.Net2MatIndex) ;
    }

  if ( HG.NrEdges > 0 ) 
    {
      if ( HG.VtxAdjncy != NULL )
	free(HG.VtxAdjncy) ;
      if ( HG.NetAdjncy != NULL )
	free(HG.NetAdjncy) ;
      if ( HG.EdgeWeight != NULL )
	free(HG.EdgeWeight) ;
      if ( HG.MatReValue != NULL )
	free(HG.MatReValue) ;
      if ( HG.MatImValue != NULL )
	free(HG.MatImValue) ;
    }

  ClearGainBucket(&(HG.GBVtx[0])) ;
  ClearGainBucket(&(HG.GBVtx[1])) ;

  HG.NrVertices = HG.NrNets = HG.NrEdges = 0 ;
  HG.CurComm = HG.MinComm = HG.OptComm = 0 ;
  HG.CurVtxLog = HG.MinVtxLog = 0 ;
  HG.SplitDir = ROW ;
}


struct biparthypergraph SparseMatrix2BiPartHyperGraph(struct sparsematrix A, 
						      int dir)
{
  register long t, tt ;
  struct biparthypergraph HG ;  
  long NrNeRows, NrNeCols ; /* Number of nonempty rows / columns  */
  long v, n, edgeindex ;
  long *nnzR, *nnzC ;
  short *Dummy, *Diag ;

  /*## Allocate memory: ##*/
  nnzR = (long *) malloc(A.m * sizeof(long)) ;
  nnzC = (long *) malloc(A.n * sizeof(long)) ;
  if ( nnzR == NULL || nnzC == NULL )
    Abort("SparseMatrix2BiPartHyperGraph(): "
	  "Not enough memory for temp arrays") ;

  /*## Clear memory: ##*/
  for ( t = 0 ; t < A.m ; t++ )
    nnzR[t] = 0 ;
  for ( t = 0 ; t < A.n ; t++ )
    nnzC[t] = 0 ;

  Dummy =  (short *) malloc(A.m * sizeof(short)) ;
  if ( Dummy == NULL )
    Abort("SparseMatrix2BiPartHyperGraph(): "
          "Not enough memory for temp arrays") ;
  for ( t = 0 ; t < A.m ; t++ )  
    Dummy[t] = 0 ;

  Diag =  (short *) malloc(A.m * sizeof(short)) ;
  if ( Diag == NULL )
    Abort("SparseMatrix2BiPartHyperGraph(): "
          "Not enough memory for temp arrays") ;
  for ( t = 0 ; t < A.m ; t++ )
        Diag[t] = 0 ;

  /*## Count the number of elements in each row / column of matrix A. ##*/
  /*## Check whether each row contains a dummy or a regular diagonal element: ##*/
  for ( t = 0, NrNeRows = 0, NrNeCols = 0 ; t < A.NrNzElts ; t++ )
    {
      if ( nnzR[A.i[t]]++ == 0 )
	NrNeRows++ ;

      if ( nnzC[A.j[t]]++ == 0 )
	NrNeCols++ ;

      if ( A.i[t] == A.j[t] && A.NrDummies > 0 &&
           BSearchLong(A.dummy, 0, A.NrDummies-1, A.i[t]) != LONG_MIN )
        Dummy[A.i[t]] = 1;

      if ( A.i[t] == A.j[t] &&  A.m == A.n &&  
         (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' || A.MMTypeCode[3]=='H')) 
        Diag[A.i[t]] = 1;
    }

  /*## If the SplitStrategy is localratio, set the direction: ##*/
  if ( Options.SplitStrategy == LocalRatio )
    {
      if ( NrNeRows > NrNeCols )   
	dir = ROW ;
      else
	if ( NrNeCols > NrNeRows )   
	  dir = COL ;
	else
	  if ( (int) ( (2.0*rand()) / (RAND_MAX+1.0) ) == 0 )
	    dir = ROW ;
	  else
	    dir = COL ;
    }

  /*## Create new Graph: ##*/
  if ( dir == ROW )
    HG = CreateNewBiPartHyperGraph(NrNeRows, NrNeCols, A.NrNzElts,
                                   TRUE, A.MMTypeCode[2]) ;
  else
    HG = CreateNewBiPartHyperGraph(NrNeCols, NrNeRows, A.NrNzElts,
                                   TRUE, A.MMTypeCode[2]) ;

  /*## Initialise the edge weights: ##*/
  for ( t = 0 ; t < A.NrNzElts ; t++ )
    HG.EdgeWeight[t] = 1 ;


  /*## Store the split direction ##*/
  HG.SplitDir = dir ;

  if ( HG.SplitDir == ROW )
    {
      /*## Insert the nonempty rows of the matrix in the split direction
           as vertices in the bipartitehypergraph and fill Vtx2MatIndex: ##*/
      for ( t = 0, tt = 0, edgeindex = 0 ; t < A.m ; t++ )
	if ( nnzR[t] > 0 )
	  {
	    HG.V[tt].iStart = HG.V[tt].iEnd = edgeindex ;	
	    HG.V[tt].vtxwgt = nnzR[t] ;

	    if ( A.NrDummies > 0  && Dummy[t] == 1)
	      HG.V[tt].vtxwgt-- ;

            if ( A.m == A.n &&
                 (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' ||
                  A.MMTypeCode[3]=='H'))
              {
                if ( Diag[t] == 1)
                  HG.V[tt].vtxwgt += (nnzR[t] -1) ;
                else
                  HG.V[tt].vtxwgt += nnzR[t] ; 
              }

	    edgeindex += nnzR[t] ;
	    HG.Vtx2MatIndex[tt] = t ;
	    tt++ ;
	  }

      /*## Insert the nonempty columns of the matrix in the other direction
           as nets in the bipartitehypergraph and fill Net2MatIndex: ##*/
      for ( t = 0, tt = 0, edgeindex = 0 ; t < A.n ; t++ )
	if ( nnzC[t] > 0 )
	  {
	    HG.N[tt].netwgt = 1 ;
	    HG.N[tt].iStartP0 = HG.N[tt].iStartP1 = HG.N[tt].iEnd = edgeindex ;
	    edgeindex += nnzC[t] ;
	    HG.Net2MatIndex[tt] = t ;
	    tt++ ;
	  }
    }
  else
    {
      /*## Insert the nonempty columns of the matrix in the split direction
           as vertices in the bipartitehypergraph and fill Vtx2MatIndex: ##*/
      for ( t = 0, tt = 0, edgeindex = 0 ; t < A.n ; t++ )
	if ( nnzC[t] > 0 )
	  {
	    HG.V[tt].iStart = HG.V[tt].iEnd = edgeindex ;	
	    HG.V[tt].vtxwgt = nnzC[t] ;

	    if ( A.NrDummies > 0  && Dummy[t] == 1 )
	      HG.V[tt].vtxwgt-- ;

 
            if ( A.m == A.n && 
                 (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' ||
                  A.MMTypeCode[3]=='H'))
              {
                if ( Diag[t] == 1)
                  HG.V[tt].vtxwgt += (nnzC[t] -1) ;
                else
                  HG.V[tt].vtxwgt += nnzC[t] ;
              }  

	    edgeindex += nnzC[t] ;
	    HG.Vtx2MatIndex[tt] = t ;
	    tt++ ;
	  }

      /*## Insert the nonempty rows of the matrix in the other direction
           as nets in the bipartitehypergraph and fill Net2MatIndex: ##*/
      for ( t = 0, tt = 0, edgeindex = 0 ; t < A.m ; t++ )
	if ( nnzR[t] > 0 )
	  {
	    HG.N[tt].netwgt = 1 ;
	    HG.N[tt].iStartP0 = HG.N[tt].iStartP1 = HG.N[tt].iEnd = edgeindex ;
	    edgeindex += nnzR[t] ;
	    HG.Net2MatIndex[tt] = t ;
	    tt++ ;
	  }
    }

  /*## Free nnzR, Dummy, and nnzC : ##*/
  if ( A.m > 0 && nnzR != NULL )
    free(nnzR) ;
  if ( A.n > 0 && nnzC != NULL )
    free(nnzC) ;
  if ( A.m > 0 && Dummy != NULL )
    free(Dummy) ;
  if ( A.m > 0 && Diag != NULL )
    free(Diag) ;

  /*## Insert the nonzero elements of the Matrix as edges in the 
       BiPartHyperGraph and fill the adjacency lists: ##*/
  for ( t = 0 ; t < A.NrNzElts ; t++ )
    {
      if ( HG.SplitDir == ROW )
	{
	  v = BSearchLong(HG.Vtx2MatIndex, 0, NrNeRows-1, A.i[t]) ;
	  n = BSearchLong(HG.Net2MatIndex, 0, NrNeCols-1, A.j[t]) ;
	}
      else
	{
	  v = BSearchLong(HG.Vtx2MatIndex, 0, NrNeCols-1, A.j[t]) ;
	  n = BSearchLong(HG.Net2MatIndex, 0, NrNeRows-1, A.i[t]) ;
	}	  

      if (A.MMTypeCode[2] != 'P')
        HG.MatReValue[HG.V[v].iEnd] = A.ReValue[t] ; 
      if (A.MMTypeCode[2] == 'C')
        HG.MatImValue[HG.V[v].iEnd] = A.ImValue[t] ; 
      
      HG.VtxAdjncy[HG.V[v].iEnd++] = n ;
      HG.NetAdjncy[HG.N[n].iEnd++] = v ;
    }

  return(HG) ;
}


void BiPartHyperGraph2SparseMatrix(struct biparthypergraph HG, 
				   struct sparsematrix T, long pl, long pr)
{
  int P ;
  long left = T.Pstart[pl] ;
  long right = T.Pstart[pr] - 1 ;
  register long t, tt ;

  for ( t = 0 ; t < HG.NrVertices ; t++ )
    {
      P = HG.OptPartVtx[t] ;

      if ( HG.SplitDir == ROW )
	{
	  for ( tt = HG.V[t].iStart ; tt < HG.V[t].iEnd ; tt++ )
	    {
	      if ( P == 0 ) 
		{
		  T.i[left] = HG.Vtx2MatIndex[t] ;
		  T.j[left] = HG.Net2MatIndex[HG.VtxAdjncy[tt]] ;
                  if (HG.MatReValue != NULL )
		    T.ReValue[left] = HG.MatReValue[tt] ;
                  if (HG.MatImValue != NULL )
		    T.ImValue[left] = HG.MatImValue[tt] ;
		  left++ ;
		}
	      else
		{
		  T.i[right] = HG.Vtx2MatIndex[t] ;
		  T.j[right] = HG.Net2MatIndex[HG.VtxAdjncy[tt]] ;
                  if (HG.MatReValue != NULL ) 
                    T.ReValue[right] = HG.MatReValue[tt] ;
                  if (HG.MatImValue != NULL )
                    T.ImValue[right] = HG.MatImValue[tt] ;
		  right-- ;
		}
	    }
	}
      else
	{
	  for ( tt = HG.V[t].iStart ; tt < HG.V[t].iEnd ; tt++ )
	    {
	      if ( P == 0 ) 
		{
		  T.i[left] = HG.Net2MatIndex[HG.VtxAdjncy[tt]] ; 
		  T.j[left] = HG.Vtx2MatIndex[t] ;
                  if (HG.MatReValue != NULL ) 
                    T.ReValue[left] = HG.MatReValue[tt] ;
                  if (HG.MatImValue != NULL )
                    T.ImValue[left] = HG.MatImValue[tt] ;
		  left++ ;
		}
	      else
		{
		  T.i[right] = HG.Net2MatIndex[HG.VtxAdjncy[tt]] ;
		  T.j[right] = HG.Vtx2MatIndex[t] ;
                  if (HG.MatReValue != NULL )  
                    T.ReValue[right] = HG.MatReValue[tt] ; 
                  if (HG.MatImValue != NULL ) 
                    T.ImValue[right] = HG.MatImValue[tt] ; 
		  right-- ;
		}
	    }
	}
    }
  
  /*## update Pstart: ##*/
  T.Pstart[(pl+pr)/2] = left ;

}

void WriteISxJS(struct biparthypergraph HG, long pl, long pr) 
{
/*Writes the indices of the rows and columns defining
  the submatrix of processors pl and pl+1, if pr=pl+2.
  Some rows or columns may be empty. */
  register long t, tt ;
  int P ;


  if ( pr - pl == 2 )
    /*Write submatrix after the last split */
    {
      if ( HG.SplitDir == ROW )
        {
          printf("I(%ld): \n",  pl+1 );
          for ( t = 0 ; t < HG.NrVertices ; t++ )
            {
	      P = HG.OptPartVtx[t] ;
	      if ( P == 0 ) 
	        printf("%ld \n", HG.Vtx2MatIndex[t]+1 ) ;
            }
          printf("J(%ld): \n",  pl+1 );
          for ( tt = 0 ; tt < HG.NrNets ; tt++ )
            printf("%ld \n", HG.Net2MatIndex[tt]+1 ) ;
          printf("I(%ld): \n",  pl+2 );
          for ( t = 0 ; t < HG.NrVertices ; t++ )
            {
	      P = HG.OptPartVtx[t] ;
	      if ( P == 1 ) 
	        printf("%ld \n", HG.Vtx2MatIndex[t]+1 ) ; 
            }
          printf("J(%ld): \n",  pl+2 );
          for ( tt = 0 ; tt < HG.NrNets ; tt++ )
            printf("%ld \n", HG.Net2MatIndex[tt]+1 ) ;
        }
      else
        {
          printf("I(%ld): \n",  pl+1 );
          for ( tt = 0 ; tt < HG.NrNets ; tt++ )
            printf("%ld \n", HG.Net2MatIndex[tt]+1 ) ;
          printf("J(%ld): \n",  pl+1 );
          for ( t = 0 ; t < HG.NrVertices ; t++ )
            {
              P = HG.OptPartVtx[t] ;
              if ( P == 0 )
                printf("%ld \n", HG.Vtx2MatIndex[t]+1 ) ;
            }
          printf("I(%ld): \n",  pl+2 );
          for ( tt = 0 ; tt < HG.NrNets ; tt++ ) 
            printf("%ld \n", HG.Net2MatIndex[tt]+1 ) ;
          printf("J(%ld): \n",  pl+2 );
          for ( t = 0 ; t < HG.NrVertices ; t++ )  
            {
              P = HG.OptPartVtx[t] ;
              if ( P == 1 ) 
                printf("%ld \n", HG.Vtx2MatIndex[t]+1 ) ;
            }
        }
    }
  fflush(stdout) ;
}
