#ifndef __HKLFM_h__
#define __HKLFM_h__

#include "IO.h"
#include "Graph.h"
#include "GainBucket.h"
#include "SparseMatrix.h"

typedef struct opts
{
  int argc ;
  char **argv ;

  /*## General options: ##*/
  char *matrix ; 
  int P ;
  double eps ;

  /*## Split strategy options: ##*/
  enum {Dynamic, Increase, Decrease} LoadbalanceStrategy ; 
  enum {Alternate, LocalBest, LocalRatio, OneDimRow, OneDimCol} SplitStrategy ;
  enum {Row, Col, Ratio} Alternate_FirstDirection ; 

  /*## Matrix options: ##*/
  enum {EqVecNo, EqVecYes} SquareMatrix_DistributeVectorsEqual ;
  enum {DumNo, DumYes} SquareMatrix_DistributeVectorsEqual_AddDummies ;
  enum {SingleEntNo, SingleEntYes} SymmetricMatrix_UseSingleEntry ;
  enum {ETypeLower, ETypeRandom} SymmetricMatrix_SingleEntryType ;

  /*## Coarsening options: ##*/
  int Coarsening_NrVertices ;
  int Coarsening_MaxNrVtxInMatch ;
  double Coarsening_StopRatio ; /* ratio CHG / HG   */
  double Coarsening_HVtxMaxFractionOfWeight ; 
  enum {Random, Inproduct} Coarsening_MatchingStrategy ;

  /*## Initial partitioning options: ##*/
  int FM_InitPart_NrRestarts ;
  int FM_InitPart_MaxNrLoops ;
  int FM_InitPart_MaxNrNoGainMoves ;

  /*## Refinement options: ##*/
  int FM_Refine_MaxNrLoops ;
  int FM_Refine_MaxNrNoGainMoves ;

  /*## Vector options: ##*/
  enum {Increasing3, Decreasing3, Random3} VectorPartition_Step3 ;

  long Seed ;

} Opts ;


void CreateInitialBalancedPartition(struct biparthypergraph *pHG) ;

void ComputeInitialGains(struct biparthypergraph *pHG) ;

void ClearMoveLog(struct biparthypergraph *pHG) ;

void MoveVertex(struct biparthypergraph *pHG, long vtx) ;

void UpdateGains(struct biparthypergraph *pHG, long vtx) ;

void RunHKLFM(struct biparthypergraph *pHG, double eps) ;

void RefineHKLFM(struct biparthypergraph *pHG, double eps) ;

void HKLFM(struct biparthypergraph *pHG, double eps, int MaxNrLoops, 
	   int MaxNrNoGainMoves) ;

void FindMatchRM(struct biparthypergraph *pHG, struct contraction *pC, long v,
                 int *Matched);
 
void FindMatchIM(struct biparthypergraph *pHG, struct contraction *pC, long v,
                 int *Matched, long *Visited, long *Inprod);

struct contraction CoarsenGraph(struct biparthypergraph *pHG,
				struct biparthypergraph *pCHG) ;

void RunMLGraphPart(struct biparthypergraph *pHG, double eps) ;

#endif /* __HKLFM_h__ */
