#include "Sort.h"

long *CSort(long *X, long LengthX, long MaxValue)
/* Sorts array X in decreasing order by counting */
{  
  register long t, ValueXt ; 
  long *C, *Xs, *I ;
 
  C  = (long *) malloc((MaxValue+1) * sizeof(long)) ;
  Xs = (long *) malloc(LengthX * sizeof(long)) ;
  I  = (long *) malloc(LengthX * sizeof(long)) ;
  if ( C == NULL || Xs == NULL || I == NULL )
    Abort("CSort(): Not enough memory for sort array !") ;

  /*## Clear memory: ##*/
  for ( t = 0 ; t <= MaxValue ; t++ )
    C[t] = 0 ;

  /*## Count value numbers: ##*/
  for ( t = 0 ; t < LengthX ; t++ )
    C[X[t]]++ ;

  /*## Create index values: ##*/
  C[0] = LengthX - C[0] ;
  for ( t = 1 ; t <= MaxValue ; t++ )
    C[t] = C[t-1] - C[t] ;

  /*## Copy the elements to the correct position: ##*/
  for ( t = 0 ; t < LengthX ; t++ )
    {
      ValueXt = X[t] ;
      I[C[ValueXt]] = t ;
      Xs[C[ValueXt]] = ValueXt ;
      C[ValueXt]++ ; 
    }

  for ( t = 0 ; t < LengthX ; t++)
    X[t] = Xs[t] ;

  if ( C != NULL )
    free(C) ;
  if ( Xs != NULL )
    free(Xs) ;

  return(I) ;
} 


/*## Binary Search value key in val[lo] - val[hi] ##*/
long BSearchLong(long *val, long lo, long hi, long key)
{
  long med ;
  
  /*## Key not found, return LONG_MIN ##*/
  if ( lo > hi )   
    return(LONG_MIN) ;
  
  med = (lo + hi) / 2 ;
  
  if ( val[med] == key )
    return(med) ;
  
  if ( key < val[med] )
    return(BSearchLong(val, lo, med-1, key)) ;
  else
    if ( key > val[med] )
      return(BSearchLong(val, med+1, hi, key)) ;

  return(-1) ; 
}

int BSearchInt(int *val, long lo, long hi, long key)
{
  long med ;
  
  /*## Key not found, return LONG_MIN ##*/
  if ( lo > hi )   
    return(INT_MIN) ;
  
  med = (lo + hi) / 2 ;
  
  if ( val[med] == key )
    return(med) ;
  
  if ( key < val[med] )
    return(BSearchInt(val, lo, med-1, key)) ;
  else
    if ( key > val[med] )
      return(BSearchInt(val, med+1, hi, key)) ;

  return(-1) ; 
}


void PermuteLong(long *X, long lo, long hi) 
{
  register long i, tmp ;
  double range ;

  while ( hi > lo )
    {
      range = (double) (hi - lo + 1) ;
      i = lo + (long) ( (range*rand()) / (RAND_MAX+1.0) ) ;
      if (i>hi) /* by rounding perhaps */
        i = hi;

      /*## swap: X[i] <-> X[hi] ##*/
      tmp = X[hi] ;
      X[hi] = X[i] ;
      X[i] = tmp ;

      hi-- ;
    }      
}

void PermuteLongLong(long *X, long *Y, long lo, long hi) 
{
  register long i, tmp ;
  double range ;

  while ( hi > lo )
    {
      range = (double) (hi - lo + 1) ;
      i = lo + (long) ( (range*rand()) / (RAND_MAX+1.0) ) ;
      if (i>hi)
        i = hi;

      /*## swap: X[i] <-> X[hi] ##*/
      tmp   = X[hi] ;
      X[hi] = X[i] ;
      X[i]  = tmp ;

      /*## swap: Y[i] <-> Y[hi] ##*/
      tmp   = Y[hi] ;
      Y[hi] = Y[i] ;
      Y[i]  = tmp ;

      hi-- ;
    }      
}

void PermuteLongDouble(long *X, double *D, long lo, long hi) 
{
  register long i, tmp ;
  double range, dtmp ;

  while ( hi > lo )
    {
      range = (double) (hi - lo + 1) ;
      i = lo + (long) ( (range*rand()) / (RAND_MAX+1.0) ) ;
      if (i>hi)
        i = hi;

      /*## swap: X[i] <-> X[hi] ##*/
      tmp   = X[hi] ;
      X[hi] = X[i] ;
      X[i]  = tmp ;

      /*## swap: D[i] <-> D[hi] ##*/
      dtmp  = D[hi] ;
      D[hi] = D[i] ;
      D[i]  = dtmp ;

      hi-- ;
    }      
}

void PermuteLongDoubleDouble(long *X, double *D, double *E, long lo, long hi)
{
  register long i, tmp ;
  double range, dtmp ;

  while ( hi > lo )
    {
      range = (double) (hi - lo + 1) ;
      i = lo + (long) ( (range*rand()) / (RAND_MAX+1.0) ) ;
      if (i>hi)
        i = hi;

      /*## swap: X[i] <-> X[hi] ##*/
      tmp   = X[hi] ;
      X[hi] = X[i] ;
      X[i]  = tmp ;

      /*## swap: D[i] <-> D[hi] ##*/
      dtmp  = D[hi] ;
      D[hi] = D[i] ;
      D[i]  = dtmp ;

      /*## swap: E[i] <-> E[hi] ##*/
      dtmp  = E[hi] ;
      E[hi] = E[i] ;
      E[i]  = dtmp ;

      hi-- ;
    }      
}

void PermuteLongLongDouble(long *X, long *Y, double *D, long lo, long hi) 
{
  register long i, tmp ;
  double range, dtmp ;

  while ( hi > lo )
    {
      range = (double) (hi - lo + 1) ;
      i = lo + (long) ( (range*rand()) / (RAND_MAX+1.0) ) ;
      if (i>hi)
        i = hi;

      /*## swap: X[i] <-> X[hi] ##*/
      tmp   = X[hi] ;
      X[hi] = X[i] ;
      X[i]  = tmp ;

      /*## swap: Y[i] <-> Y[hi] ##*/
      tmp   = Y[hi] ;
      Y[hi] = Y[i] ;
      Y[i]  = tmp ;

      /*## swap: D[i] <-> D[hi] ##*/
      dtmp  = D[hi] ;
      D[hi] = D[i] ;
      D[i]  = dtmp ;

      hi-- ;
    }      
}
